From 4caf8049b70f04c02b9f24e6b4cae73965884e0b Mon Sep 17 00:00:00 2001
From: Alyssa Rosenzweig <alyssa@rosenzweig.io>
Date: Wed, 5 Jul 2023 12:19:51 -0400
Subject: [PATCH] asahi,agx: Set coherency bit for clustered targets

We need to set a particular bit on atomics for them to be coherent across
clusters. Fixes atomics on G13X.

Setting this bit on the single-cluster G13G, on the other hand, wedges the GPU.
So best be careful ;-)

Signed-off-by: Alyssa Rosenzweig <alyssa@rosenzweig.io>
---
 src/asahi/compiler/agx_compile.h      |  3 +++
 src/asahi/compiler/agx_pack.c         | 10 ++++++----
 src/gallium/drivers/asahi/agx_state.c |  4 +++-
 3 files changed, 12 insertions(+), 5 deletions(-)

diff --git a/src/asahi/compiler/agx_compile.h b/src/asahi/compiler/agx_compile.h
index a936e9a91b5..713ebdf6f06 100644
--- a/src/asahi/compiler/agx_compile.h
+++ b/src/asahi/compiler/agx_compile.h
@@ -194,6 +194,9 @@ struct agx_shader_key {
    /* Number of reserved preamble slots at the start */
    unsigned reserved_preamble;
 
+   /* Does the target GPU have multiple clusters? Affects cache policy */
+   bool clustered;
+
    union {
       struct agx_vs_shader_key vs;
       struct agx_fs_shader_key fs;
diff --git a/src/asahi/compiler/agx_pack.c b/src/asahi/compiler/agx_pack.c
index c5ab5da1a98..085447431d0 100644
--- a/src/asahi/compiler/agx_pack.c
+++ b/src/asahi/compiler/agx_pack.c
@@ -512,7 +512,7 @@ agx_pack_alu(struct util_dynarray *emission, agx_instr *I)
 
 static void
 agx_pack_instr(struct util_dynarray *emission, struct util_dynarray *fixups,
-               agx_instr *I)
+               agx_instr *I, bool coherent_across_clusters)
 {
    switch (I->op) {
    case AGX_OPCODE_LD_TILE:
@@ -736,8 +736,10 @@ agx_pack_instr(struct util_dynarray *emission, struct util_dynarray *fixups,
          (I->scoreboard << 30) |
          (((uint64_t)((O >> 4) & BITFIELD_MASK(4))) << 32) |
          (((uint64_t)((A >> 4) & BITFIELD_MASK(4))) << 36) |
-         (((uint64_t)(R >> 6)) << 40) | (Rt ? BITFIELD64_BIT(47) : 0) |
-         (((uint64_t)S) << 48) | (((uint64_t)(O >> 8)) << 56);
+         (((uint64_t)(R >> 6)) << 40) |
+         (coherent_across_clusters ? BITFIELD64_BIT(45) : 0) |
+         (Rt ? BITFIELD64_BIT(47) : 0) | (((uint64_t)S) << 48) |
+         (((uint64_t)(O >> 8)) << 56);
 
       memcpy(util_dynarray_grow_bytes(emission, 1, 8), &raw, 8);
       break;
@@ -994,7 +996,7 @@ agx_pack_binary(agx_context *ctx, struct util_dynarray *emission)
       block->offset = emission->size;
 
       agx_foreach_instr_in_block(block, ins) {
-         agx_pack_instr(emission, &fixups, ins);
+         agx_pack_instr(emission, &fixups, ins, ctx->key->clustered);
       }
    }
 
diff --git a/src/gallium/drivers/asahi/agx_state.c b/src/gallium/drivers/asahi/agx_state.c
index c7c59c0cf38..05d6758e993 100644
--- a/src/gallium/drivers/asahi/agx_state.c
+++ b/src/gallium/drivers/asahi/agx_state.c
@@ -1665,7 +1665,9 @@ agx_compile_variant(struct agx_device *dev, struct agx_uncompiled_shader *so,
       }
    }
 
-   struct agx_shader_key base_key = {0};
+   struct agx_shader_key base_key = {
+      .clustered = dev->params.num_clusters_total > 1,
+   };
 
    if (nir->info.stage == MESA_SHADER_FRAGMENT)
       base_key.fs.nr_samples = key_->fs.nr_samples;
-- 
2.39.2

